
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%% Corecursive Featherweight Java Prolog implementation
%% Please refer to file tests.pl for execution tests.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


%%:- module(coFJ_translation,[evaluate/2]).

:- use_module(meta_interpreter).

%% Main:

evaluate(E,V) :- solve(eval(E,V)).


%% Utilities:

eq(X,Y,true) :- X==Y,!.
eq(_,_,false).

neq(X,Y,false) :- X==Y,!.
neq(_,_,true).

lth(X,Y,true) :- X<Y,!.
lth(_,_,false).

gth(X,Y,true) :- X>Y,!.
gth(_,_,false).

leq(X,Y,true) :- X=<Y,!.
leq(_,_,false).

geq(X,Y,true) :- X>=Y,!.
geq(_,_,false).

and(X,Y,true) :- X,Y,!.
and(_,_,false).

or(X,Y,true) :- (X;Y),!.
or(_,_,false).

not(X,true) :- \+X,!.
not(_,false).

neg(X,-X).



%% CoFJ Translation:


%% The predicate fields/2 takes as arguments class name and list of fields. 
%% For instance, if we consider Object, we have fields(object,[]).  
%% Moreover, a derived predicate field/3 takes as arguments class name, field name and index, 
%% and is used to retrieve the position of a field in the list of fields of the given class.
%% This is useful since a field access looks for such position, starting from the field name.

%% fields/2: (receiver, fields)

fields(object,[]).
fields(list,[]).
fields(eList,[]).
fields(cycListFact,[]).
fields(nEList,[el,nx]).


%% field/3: (receiver,field,position)

field(C,F,I) :- fields(C,FL), nth0(I,FL,F).



%% The predicate mbody/5 takes as arguments class name, method name, implicit This parameter, list or parameters, and definition. 
%% The predicate co-mbody/5 is the analogous for the codefinition.

%% mbody/5: (class_name,method_name,receiver,aruments,body)
%% co-mbody/5: (class_name,method_name,receiver,aruments,codefinition)


%% Empty List definitions:
mbody(eList,member,_This,[_X],false).
mbody(eList,allPos,_This,[],true).
mbody(eList,allNeg,_This,[],true).
mbody(eList,isCyc,_This,[],false).


%% Non-Empty List definitions:
mbody(nEList,member,This,[X],E) :- E = if(eq(field_acc(This,el),X),true,invoke(field_acc(This,nx),member,[X])).
mbody(nEList,allPos,This,[],E) :- E = if(lth(field_acc(This,el),0),false,invoke(field_acc(This,nx),allPos,[])).
mbody(nEList,allNeg,This,[],E) :- E = if(geq(field_acc(This,el),0),false,invoke(field_acc(This,nx),allNeg,[])).
mbody(nEList,isCyc,This,[],E) :- E = invoke(field_acc(This,nx),isCyc,[]).
mbody(cycListFact,infOcc,This,[X],E) :- E = new(nEList,[X,invoke(This,infOcc,[X])]).

%% Codefinitions:  
co-mbody(nEList,member,_This,[_X],false).
co-mbody(nEList,allPos,_This,[],true).
co-mbody(nEList,allNeg,_This,[],true).
co-mbody(nEList,isCyc,_This,[],true).
co-mbody(cycListFact,infOcc,_This,[_X],any).



%% The auxiliary predicate isValue/1 holds if the argument is a value, that is, 
%% either an integer or an object. Predicates isValue/1 and isValueList/1 are defined as follows:

%% isValue/1:

%% Used to evaluate more than one expression
isValueList([]).
isValueList([H|T]) :- isValue(H), isValueList(T).

isValue(new(_,FVL)) :- isValueList(FVL).
isValue(X) :- integer(X).



%% Core Semantics:


%% eval\2: (Exp, ResValue)

evalList([],[]).
evalList([H|T], [H1|T1]) :- eval(H,H1), evalList(T,T1).

%% Corresponds to rule abs-new
eval(new(C,Args),new(C,VL)) :- evalList(Args,VL).

%% Corresponds to rule abs-field   
eval(field_acc(TO,F),V) :- eval(TO,new(C,FVL)), field(C,F,I), nth0(I,FVL,V).

%% Corresponds to rule abs-invk. Note that parameter passing is modelled via unification. 
eval(invoke(TO,M,Args),RV) :- eval(TO,new(C,FLV)), evalList(Args,VArgs), mbody(C,M,new(C,FLV),VArgs,E), eval(E,RV). 


%% Utilities:

%% The meaning of the clause eval(any,_) is that any evaluates to anything. 
%% For simplicity, in this implementation we decided to treat it as a constant, 
%% differently from the abstract semantics where it is a variable.

eval(any,_).

%% eval(X,X) handles values, and corresponds to corule \refToRule{abs-co-val} in the abstract semantics.
%% Values are handled by a rule rather than a corule to uniformly handle also primitive values.

eval(X,X) :- isValue(X).

eval(true,true).
eval(false,false).

eval(if(G,T,_),RV) :- eval(G,true), eval(T,RV).
eval(if(G,_,F),RV) :- eval(G,false), eval(F,RV).
eval(eq(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), eq(V1,V2,RV). 
eval(lth(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), lth(V1,V2,RV). 
eval(leq(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), leq(V1,V2,RV).
eval(gth(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), gth(V1,V2,RV).
eval(geq(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), geq(V1,V2,RV).
eval(and(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), and(V1,V2,RV). 
eval(or(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), or(V1,V2,RV).
eval(not(E1),RV) :- eval(E1,V1), not(V1,RV). 
eval(neg(E1),RV) :- eval(E1,V1), neg(V1,RV).
eval(sum(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), RV is V1 + V2.
eval(dif(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), RV is V1 - V2.
eval(prd(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), RV is V1 * V2.
eval(div(E1,E2),RV) :- eval(E1,V1), eval(E2,V2), RV is V1 / V2.


%% Corules:

%% Used to handle cyclic objects
co(isValue(_)).

%% Corresponds to corule abs-co-invk. Parameter passing is again modelled via unification. 
co(eval(invoke(TO,M,Args),RV)) :- eval(TO,new(C,FLV)), evalList(Args,VArgs), co-mbody(C,M,new(C,FLV),VArgs,E), eval(E,RV). 

